/**
 * 传统版Ajax(基于XMLHttpRequest)
 * @author yutent<yutent.io@gmail.com>
 * @date 2020/08/03 17:05:10
 */

import { Format, getType } from './lib/format.js'

const NOBODY_METHODS = ['GET', 'HEAD']
const FORM_TYPES = {
  form: 'application/x-www-form-urlencoded; charset=UTF-8',
  json: 'application/json; charset=UTF-8',
  text: 'text/plain; charset=UTF-8'
}
const ERRORS = {
  10001: 'Argument url is required',
  10012: 'Parse error',
  10100: 'Request canceled',
  10104: 'Request pending...',
  10200: 'Ok',
  10204: 'No content',
  10304: 'Not modified',
  10500: 'Internal Server Error',
  10504: 'Connected timeout'
}

Promise.defer = function () {
  let _ = {}
  _.promise = new Promise(function (y, n) {
    _.resolve = y
    _.reject = n
  })
  return _
}

class _Request {
  constructor(url = '', options = {}, owner) {
    if (!url) {
      throw new Error(ERRORS[10001])
    }

    // url规范化
    url = url.replace(/#.*$/, '')

    if (owner.BASE_URL) {
      if (!/^([a-z]+:|\/\/)/.test(url)) {
        url = owner.BASE_URL + url
      }
    }

    options.method = (options.method || 'get').toUpperCase()

    this.xhr = new XMLHttpRequest()
    this.defer = Promise.defer()
    this._owner = owner

    this.options = {
      headers: {
        'content-type': FORM_TYPES.form
      },
      body: null,
      cache: 'default',
      credentials: false, // 跨域选项,是否验证凭证
      signal: null, // 超时信号, 配置该项时, timeout不再生效
      timeout: 30000 // 超时时间, 单位毫秒, 默认30秒
    }

    if (!options.signal) {
      let control = new AbortController()
      options.signal = control.signal
    }
    this.defer.promise.abort = function () {
      control.abort()
    }

    if (options.headers) {
      let headers = this.options.headers
      Object.assign(headers, options.headers)
      options.headers = headers
    }

    Object.assign(this.options, options, { url })

    if (owner._inject_req) {
      owner._inject_req(this.options)
    }

    this.__next__()
    return this.defer.promise
  }

  __next__() {
    let options = this.options
    let params = null
    let hasAttach = false // 是否有附件
    let noBody = NOBODY_METHODS.includes(options.method)

    /* ------------------------  1»» 处理signal  ---------------------- */
    options.signal.onabort = _ => {
      this.cancel = true
      this.xhr.abort()
    }

    /* --------------------------  2»» 请求的内容 --------------------- */
    if (options.body) {
      let type = typeof options.body
      switch (type) {
        case 'number':
        case 'string':
          this.__type__('text')
          params = options.body
          break
        case 'object':
          let _type = getType(options.body)

          if (_type === 'ArrayBuffer' || _type === 'Uint8Array') {
            break
          }
          // 解析表单DOM
          if (options.body.nodeName === 'FORM') {
            options.method = options.body.method.toUpperCase() || 'POST'

            params = Format.parseForm(options.body)
            hasAttach = params.constructor === FormData

            // 如果是一个 FormData对象,且为不允许携带body的方法,则直接改为POST
          } else if (options.body.constructor === FormData) {
            hasAttach = true
            if (noBody) {
              options.method = 'POST'
            }
            params = options.body
          } else {
            for (let k in options.body) {
              if (Array.isArray(options.body[k]) || getType(options.body[k]) === 'FileList') {
                options.body[k] = Array.from(options.body[k])
                hasAttach = options.body[k].some(
                  it => getType(it) === 'File' || getType(it) === 'Blob'
                )
              } else {
                if (getType(options.body[k]) === 'File' || getType(options.body[k]) === 'Blob') {
                  hasAttach = true
                  break
                }
              }
            }

            // 有附件,则改为FormData
            if (hasAttach) {
              if (noBody) {
                options.method = 'POST'
              }
              params = Format.mkFormData(options.body)
            } else {
              params = options.body
            }
          }
      }
    } else {
      delete options.headers['content-type']
    }

    if (hasAttach) {
      delete options.headers['content-type']
    }

    /* --------------------------  3»» 处理跨域  --------------------- */
    try {
      let crossDomain
      let $a = document.createElement('a')

      $a.href = options.url
      crossDomain = location.protocol !== $a.protocol || location.host !== $a.host
      $a = null

      if (crossDomain) {
        if (options.credentials) {
          this.xhr.withCredentials = true
        }
      }
    } catch (err) {}

    /* ------------- 4»» 根据method类型, 处理表单数据  ---------------- */

    // 拼接到url上
    if (noBody) {
      params = Format.param(params)
      if (params) {
        options.url += (options.url.includes('?') ? '&' : '?') + params
      }
      if (options.cache === 'no-store') {
        options.url += (options.url.includes('?') ? '&' : '?') + '_t_=' + Date.now()
      }
    } else {
      if (options.body && !hasAttach) {
        if (options.headers['content-type'].includes('json')) {
          params = JSON.stringify(params)
        }
        if (options.headers['content-type'].includes('form')) {
          params = Format.param(params)
        }
      }
    }

    /* -----------------     5»» 设置响应的数据类型   ---------------- */
    // 统一使用blob, 再转为其他的
    this.xhr.responseType = 'blob'

    /* -----------------    6»» 构造请求    ------------------- */
    // 6.1
    this.xhr.onreadystatechange = ev => {
      if (options.timeout > 0) {
        this['time' + this.xhr.readyState] = ev.timeStamp
        if (this.xhr.readyState === 4) {
          options.isTimeout = this.time4 - this.time1 > options.timeout
        }
      }

      if (this.xhr.readyState !== 4) {
        return
      }
      this.__dispatch__(options.isTimeout)
    }

    // 6.2»» 初始化xhr
    this.xhr.open(options.method, options.url)

    // 6.3»» 设置头信息
    for (let k in options.headers) {
      this.xhr.setRequestHeader(k, options.headers[k])
    }

    // 6.4»» 发起网络请求
    this.xhr.send(params)

    // 6.5»» 超时处理
    if (options.timeout && options.timeout > 0) {
      this.xhr.timeout = options.timeout
    }
  }

  __type__(type) {
    this.options.headers['content-type'] = FORM_TYPES[type]
  }

  __dispatch__(isTimeout) {
    let result = {
      status: 200,
      statusText: 'ok',
      body: '',
      headers: Object.create(null)
    }

    // 主动取消
    if (this.cancel) {
      return this.__cancel__()
    }

    // 超时
    if (isTimeout) {
      return this.__timeout__()
    }

    // 是否请求成功(resful规范)
    let isSucc = this.xhr.status >= 200 && this.xhr.status < 400
    let headers = this.xhr.getAllResponseHeaders().split('\n') || []

    //处理返回的 Header, 拿到content-type
    for (let it of headers) {
      it = it.trim()
      if (it) {
        it = it.split(':')
        let k = it.shift().toLowerCase()
        it = it.join(':').trim()
        result.headers[k] = it
      }
    }

    if (isSucc) {
      result.status = this.xhr.status
      if (result.status === 204) {
        result.statusText = ERRORS[10204]
      } else if (result.status === 304) {
        result.statusText = ERRORS[10304]
      }
    } else {
      result.status = this.xhr.status || 500
      result.statusText = this.xhr.statusText || ERRORS[10500]
    }

    result.body = result.status === 204 ? null : this.xhr.response

    this.__success__(isSucc, result)
  }

  __success__(isSucc, result) {
    let { body, status, statusText, headers } = result
    let response = new Response(body, { status, statusText, headers })
    let _type

    if (this._owner._inject_res) {
      response = this._owner._inject_res(response)
      _type = getType(it)(response)
    }

    if (isSucc) {
      this.defer.resolve(response)
    } else {
      if (_type === 'Promise') {
        return response.then(_ => this.defer.reject(_)).catch(_ => this.defer.reject(_))
      } else {
        this.defer.reject(response)
      }
    }
    delete this.xhr
    delete this.options
    delete this.defer
  }

  __cancel__(result) {
    let response = new Response('', {
      status: 0,
      statusText: ERRORS[10100]
    })

    this.defer.reject(response)

    delete this.xhr
    delete this.options
    delete this.defer
  }

  __timeout__(result) {
    let response = new Response('', {
      status: 504,
      statusText: ERRORS[10504]
    })

    this.defer.reject(response)

    delete this.xhr
    delete this.options
    delete this.defer
  }
}

function inject(target) {
  target.inject = {
    request(callback) {
      target._inject_req = callback
    },
    response(callback) {
      target._inject_res = callback
    }
  }
}
const _fetch = function (url, options) {
  return new _Request(url, options, _fetch)
}

_fetch.create = function () {
  let another = function (url, options) {
    return new _Request(url, options, another)
  }
  inject(another)
  return another
}

inject(_fetch)

export default _fetch
